#include "blob.h"
#include "raytrace.h"
#include "ray.h"
#include <cmath>
#include <map>
using namespace std;

// fonction polynomiale du second degr dfinie par ses trois cofficients a * x^2 + b * x + c
struct poly
{
    float a, b, c;
};

const int zoneNumber = 10;

// une zone est dfinie par un rapport du carr de la distance  la sphre sur le carr de son rayon
// puis par les gamma et beta incrmentaux qui approchent la courbe 1/ dist^2
// ces deux coefficients ne sont calcules qu'une seule fois par excution dans l'initZones.
struct
{
    float fCoef, fGamma, fBeta;
} zoneTab[zoneNumber] = 
{   
    {10.0f, 0, 0},
    {5.0f,  0, 0},
    {3.33333f,  0, 0},
    {2.5f,  0, 0},
    {2.0f,  0, 0},
    {1.66667f,  0, 0},
    {1.42857f,  0, 0},
    {1.25f,  0, 0},
    {1.1111f,  0, 0},
    {1.0f,  0, 0} 
};

static void initZones()
{
    float fLastGamma = 0.0f, fLastBeta = 0.0f;
    float fLastInvRSquare = 0.0f;
    for (int i = 0; i < zoneNumber - 1; i++)
    {
        float fInvRSquare = 1.0f / zoneTab[i + 1].fCoef;
        // fGamma est la pente entre le point d'entre et le point de sortie
        // on ne stocke que la valeur incrmentale par rapport  la zone prcdente
        zoneTab[i].fGamma = (fLastInvRSquare - fInvRSquare) / (zoneTab[i].fCoef - zoneTab[i + 1].fCoef) - fLastGamma;
        fLastGamma = zoneTab[i].fGamma + fLastGamma;
        // en faisant - fLastGamma et + fLastGamma  la ligne suivante 
        // on s'conomise l'utilisation d'une temporaire
        // je sais c'est dbile..

        // fBeta est la valeur de la droite approchant la courbe pour dist = 0
        // on ne stocke que la valeur incrmentale par rapport  la zone prcdente
        zoneTab[i].fBeta = (fInvRSquare - fLastGamma * zoneTab[i+1].fCoef) - fLastBeta;
        fLastBeta = zoneTab[i].fBeta + fLastBeta;
        fLastInvRSquare = fInvRSquare;
    };
    // la derniere zone d'influence agit comme un signal que l'on traite comme un cas particulier
    zoneTab[zoneNumber - 1].fGamma = 0.0f;
    zoneTab[zoneNumber - 1].fBeta = 0.0f;
}

bool isBlobIntersected(const ray &r, const blob& b, float &t)
{
    map<float, poly> polynomMap;
    static bool bInit = false;

    if (!bInit)
    {
        // on initialise les cofficients gamma et beta une seule fois par execution
        initZones();
        bInit = true;
    }
    
    // tous ces calculs pourraient probablement 
    // etre rduit lors d'un prcalcul mais j'essaie
    // de privilgier la clart  la rapidit pour l'instant

    float rSquare, rInvSquare;
    rSquare = b.size * b.size;
    // Cela ne sert  rien d'avoir des spheres de taille zero
    // mais soyons de bons paranoiaques tout de meme
    if (rSquare == 0.0f)
        return false;
    rInvSquare = 1.0f / rSquare;

    for (unsigned int i= 0; i< b.centerList.size(); i++)
    {
        float A, B, C;
        float fDelta, t0, t1;
        point currentPoint = b.centerList[i];

    	vecteur vDist = currentPoint - r.start;
        A = 1.0f;
	    B = - 2.0f * r.dir * vDist;
        C = vDist * vDist;

        // on parcourt la liste des zones d'influences de la sphre courante
        // on calcule la nouvelle version du polynome qui a cours dans 
        // cette zone d'influence et on le stocke de manire incrmentale
        // ce qui importe c'est la diffrence avec la zone prcdente, ce qui permet
        // de bien grer le cas de sphres imbriques les unes dans les autres
        for (int j=0; j < zoneNumber - 1; j++)
        {
            // On calcule le Delta de l'quation s'il est ngatif
            // il n'y a pas de solution donc pas de point d'intersection
            fDelta = B*B - 4.0f * (C - zoneTab[j].fCoef * rSquare);
            if (fDelta < 0.0f) 
                break;
            t0 =  0.5f * ( - B - sqrtf(fDelta)); 
            // cool on ne s'occupe pas de l'ordre il est ici explicite t0 < t1
	        t1 =  0.5f * ( - B + sqrtf(fDelta));
            poly poly0 = {zoneTab[j].fGamma * A * rInvSquare ,
                          zoneTab[j].fGamma * B * rInvSquare , 
                          zoneTab[j].fGamma * C * rInvSquare + zoneTab[j].fBeta }; 
            poly poly1 = {- poly0.a, - poly0.b, - poly0.c};
            
            // les variations du polynome sont tries par leur position sur le rayon
            // au fur et  mesure qu'elles sont insres. C'est la map qui nous garantit cela.
            // ce serait peut-etre plus optimal de placer dans un vector sans se soucier de l'ordre
            // et trier ensuite mais je me fiche de l'optimisation en fait.
            polynomMap.insert(pair<float, poly>(t0, poly0));
            polynomMap.insert(pair<float, poly>(t1, poly1));
        };
    }

    // en dehors de toute zone d'influence le champ de potentiel est nul
    poly currentPolynom = {0.0f,0.0f,0.0f};

    for (map<float, poly>::const_iterator it = polynomMap.begin(); it != polynomMap.end(); ++it  )
    {
        // comme on a stock les polynomes de manire incrmentale on 
        // reconstruit le polynome de la zone d'influence courante
        currentPolynom.a += (*it).second.a;
        currentPolynom.b += (*it).second.b;
        currentPolynom.c += (*it).second.c;

        map<float, poly>::const_iterator itNext = it;
        ++itNext;

        // a ne sert  rien de rsoudre l'quation si la zone d'influence est avant le point de dpart
        // ou aprs le point d'arrive sur le rayon
        if ((t > (*it).first ) && ((*itNext).first) > 0.01f)
        {
            // on peut se permettre de rsoudre la dernire quation de manire exacte
            // aprs toutes les approximations que l'on a fait
            // avec un nombre suffisant de dcoupages,
            // il devrait tre difficile de faire la distinction
            // entre le blob et son dcoupage approch.
            float fDelta = currentPolynom.b * currentPolynom.b 
                - 4.0f * currentPolynom.a * (currentPolynom.c - 1.0f) ;
            if (fDelta < 0.0f)
                continue;

            bool retValue = false;

            float t0 = (0.5f / currentPolynom.a) * (- currentPolynom.b - sqrtf(fDelta)); 
            if ((t0 > 0.01f ) && (t0 >= (*it).first ) && (t0 < (*itNext).first) && (t0 <= t ))
            {
                t = t0;
                retValue = true;
            }
            float t1 = (0.5f / currentPolynom.a) * (- currentPolynom.b + sqrtf(fDelta));
            if ((t1 > 0.01f ) && (t1 >= (*it).first ) && (t1 < (*itNext).first) && (t1 <= t ))
            {
                t = t1;
                retValue = true;
            }

            if (retValue)
                return true;        
        }
    }

    return false;
}

void blobInterpolation(point pos, const blob& b, const material matList[], vecteur &vOut, material& matOut)
{
    vecteur gradient = {0.0f,0.0f,0.0f};
    // float fSomme = 0.0f;
    float fRSquare = b.size * b.size;
    for (unsigned int i= 0; i< b.centerList.size(); i++)
    {
        vecteur normal = pos - b.centerList[i];
        float fDistSquare = normal * normal;
        if (fDistSquare <= 0.001f) 
            continue;
        float fDistFour = fDistSquare * fDistSquare;
        normal = (fRSquare/fDistFour) * normal;

        // c'est la vraie formule du gradient dans un champ de potentiel 
        // et non pas une simple moyenne
        gradient = gradient + normal;
        // fSomme = fSomme + (fRSquare/fDistFour);
    }
    vOut = gradient;
}
