#include <iostream>
#include <fstream>
#include <vector>
#include <cmath>
#include <limits>
#include <algorithm>
using namespace std;

#include "Ray.h"
#include "Raytrace.h"
#include "SimpleString.h"
#include "Texture.h"
#include "Perlin.h"
#include "Scene.h"
#include "Sphere.h"
#include "Photonmap.h"
#include "Photontrace.h"

float addRay(ray viewRay, scene &myScene, context myContext, photonmap & myPhotonMap)
{
    float output = 0.0f; 
    float coef = 1.0f;
    int level = 0;

    do 
    {
        point ptHitPoint;
        vecteur vNormal;
        material currentMat;
        {
            float distance;
            if (!findIntersection(/*IN*/viewRay, /*IN*/myScene, 
                                /*OUT*/ptHitPoint, /*OUT*/vNormal, 
                                /*OUT*/currentMat, /*OUT*/distance))
            {
                break;
            }

            if (myContext.absorption != 1.0f)
            {
                // la lumire transmise dans le mdium transparent
                // est affecte par la diffusion due aux impurets
                // la beer's law s'applique pour les faibles concentrations
                coef = coef * powf(myContext.absorption, distance);
            }
        }
        
        bool bInside;

        if (vNormal * viewRay.dir > 0.0f)
        {
            vNormal = -1.0f * vNormal;
            bInside = true;
        }
        else
        {
            bInside = false;
        }

        if (currentMat.bump)
        {
            float noiseCoefx = float(noise(0.1 * double(ptHitPoint.x), 0.1 * double(ptHitPoint.y),0.1 * double(ptHitPoint.z)));
            float noiseCoefy = float(noise(0.1 * double(ptHitPoint.y), 0.1 * double(ptHitPoint.z),0.1 * double(ptHitPoint.x)));
            float noiseCoefz = float(noise(0.1 * double(ptHitPoint.z), 0.1 * double(ptHitPoint.x),0.1 * double(ptHitPoint.y)));
            
            vNormal.x = (1.0f - currentMat.bump ) * vNormal.x + currentMat.bump * noiseCoefx;  
            vNormal.y = (1.0f - currentMat.bump ) * vNormal.y + currentMat.bump * noiseCoefy;  
            vNormal.z = (1.0f - currentMat.bump ) * vNormal.z + currentMat.bump * noiseCoefz;  
            
            float temp = vNormal * vNormal;
            if (temp == 0.0f)
            break;
            temp = invsqrtf(temp);
            vNormal = temp * vNormal;
        }
        const materialChannel & currentMatChannel = currentMat.getChannel(myContext.offset);
    
        ray outRay = viewRay;

        bool bDiffuse = LookForDiffuse(viewRay, ptHitPoint, vNormal, currentMatChannel, bInside, myContext, outRay);

        // Maintenant qu'on a dtermin quelle tait le point visible
        // sur l'cran et dans quel direction pointait la surface
        // en ce point, il reste  calculer la valeur d'clairement
        // L encore il faut parcourir la liste des lumires
        // et dterminer celles qui ont une influence sur notre objet.

        if (bDiffuse)
        {
            float noiseCoef = 0.0f;
            
            switch (currentMat.type) {
                case material::noise:
                    noiseCoef = noisef(0.1f * ptHitPoint.x, 0.1f * ptHitPoint.y,0.1f * ptHitPoint.z);
                    break;
                case material::turbulence:
                {
                    for (int level = 1; level < 10; level ++)
                    {
                        noiseCoef += (1.0f / level ) 
                            * fabsf(noisef(level * 0.05f * ptHitPoint.x, 
                                        level * 0.05f * ptHitPoint.y,
                                        level * 0.05f * ptHitPoint.z));
                    };
                    break;
                }
                case material::marble:
                {
                    for (int level = 1; level < 10; level ++)
                    {
                        noiseCoef +=  (1.0f / level) 
                            * fabsf(noisef(level * 0.05f * ptHitPoint.x, 
                                        level * 0.05f * ptHitPoint.y, 
                                        level * 0.05f * ptHitPoint.z));
                    };
                    noiseCoef = 0.5f * sinf( (ptHitPoint.x + ptHitPoint.y) * 0.05f + noiseCoef) + 0.5f;
                    break;
                }
                default:
                    break;
            }

            float fLightContrib = 0.0f;

            float fViewProjection = viewRay.dir * vNormal;

            static priorityHeap nearestPhoton;
            nearestPhoton.maxDistSquare = myScene.phMapProp.maxDistanceSquare;
            nearestPhoton.resultTab.clear();

            int found = myPhotonMap.findNearestPhotons(ptHitPoint, vNormal, 1.0f, myScene.phMapProp.maxNearestPhotons, nearestPhoton);
            
            if ( found > 0 && nearestPhoton.maxDistSquare > 0.0f)
            {
                float fInvMaxDist = invsqrtf(nearestPhoton.maxDistSquare);

                vector<pair<float, const photon * > >::const_iterator it = nearestPhoton.resultTab.begin();
                for (;it != nearestPhoton.resultTab.end() ; ++it)
                {
                    float fPhotonDist = sqrtf((*it).first); 
                    float fPhotonWeight;

                    switch (myScene.phMapProp.eFilterType)
                    {
                    case photonMapSettings::conicEstimator:
                        {
                            static const float invKappa = 0.9f;
                            fPhotonWeight =  1.0f - fPhotonDist * fInvMaxDist * invKappa;
                        }
                        break;
                    case photonMapSettings::gaussianEstimator:
                        {
                            static const float Alpha = 1.818f;
                            static const float Beta = 1.953f;
                            float fDistRatio = fPhotonDist * fInvMaxDist;
                            fPhotonWeight =  Alpha * 
                                (1.0f - (1.0f - expf(-0.5f * Beta * fDistRatio * fDistRatio)) / (1.0f - expf (-Beta) ));
                        }
                        break;
                    default: // photonMapSettings::RadianceEstimator:
                        {
                            fPhotonWeight = 1.0f;
                        }
                        break;
                    }

                    float fLightIntensity = fPhotonWeight;

                    float fLightProjection = - ((*it).second->dir * vNormal) ;

                    if (fLightProjection > 0.0f)
                    {
                        float lambert = 1.0f;
                        
                        fLightContrib += lambert * fLightIntensity 
                            * ((1.0f - noiseCoef)* currentMatChannel.diffuse + 
                                 noiseCoef * currentMatChannel.diffuse2);
                		
			            // Blinn 
                        // La direction de Blinn est exactement  mi chemin entre le rayon
                        // lumineux et le rayon de vue. 
                        // On calcule le vecteur de Blinn et on le rend unitaire
                        // puis on calcule le cofficient de blinn
                        // qui est la contribution spculaire de la lumire courante.

                        vecteur blinnDir = (*it).second->dir + viewRay.dir;
                        float temp = blinnDir * blinnDir;
                        if (temp != 0.0f )
                        {
                            float blinn = invsqrtf(temp) * max(fLightProjection - fViewProjection , 0.0f);
                            blinn = currentMatChannel.specular * powf(blinn, currentMatChannel.power);
                            fLightContrib += blinn * fLightIntensity;
                        }
                    }
                }
                if(fLightContrib > 0.0f)
                {
                    switch (myScene.phMapProp.eFilterType)
                    {
                    case photonMapSettings::conicEstimator:
                        {
                            static const float invKappa = 0.9f;
                            static const float divider = 1.0f / ( (1.0f - 0.666666666667f * invKappa) * 3.1415927f);
                            fLightContrib *= myPhotonMap.getWattagePerPhoton()
                                *  
                                divider * fInvMaxDist * fInvMaxDist  ;
                        }
                        break;
                    case photonMapSettings::gaussianEstimator:
                        {
                            const float divider = 1.0f / 3.1415927f;
                            fLightContrib *= myPhotonMap.getWattagePerPhoton()
                                          * divider * fInvMaxDist * fInvMaxDist ;
                        }
                        break;
                    default: // photonMapSettings::RadianceEstimator:
                        {
                            const float divider = 1.0f / 3.1415927f;

                            fLightContrib *= myPhotonMap.getWattagePerPhoton()
                                          * divider * fInvMaxDist * fInvMaxDist  ;                        }
                        break;
                    }

                }
            }
            if (!myScene.phMapProp.directLighting)
            {
                ray lightRay;
                lightRay.start = ptHitPoint;

                for (unsigned int j = 0; j < myScene.lgtTab.size() ; ++j)
                {
                    light currentLight = myScene.lgtTab[j];

                    lightRay.dir = currentLight.pos - ptHitPoint;
                    float fLightProjection = lightRay.dir * vNormal;
                    float fLightIntensity = currentLight.intensity.getChannel(myContext.offset);

                    if ( fLightProjection <= 0.0f )
                        continue;

                    float lightDist = lightRay.dir * lightRay.dir;
                    {
                        float temp = lightDist;
			            if ( temp == 0.0f )
				            continue;
                        temp = invsqrtf(temp);
			            lightRay.dir = temp * lightRay.dir;
                        lightDist = temp * lightDist ;
                        fLightProjection = temp * fLightProjection;
                        fLightIntensity = temp * temp * fLightIntensity * (1.0f / 4.0f / 3.1415927f);
                    }

                    bool inShadow = false;
                    {
                        float t = lightDist;
                        for (unsigned int i = 0; i < myScene.sphTab.size() ; ++i)
                        {
                            if (hitSphere(lightRay, myScene.sphTab[i], t))
                            {
                                inShadow = true;
                                break;
                            }
                        }
                        for (unsigned int i = 0; i < myScene.blbTab.size() ; ++i)
                        {
                            if (isBlobIntersected(lightRay, myScene.blbTab[i], t)) {
                                inShadow = true;
                                break;
                            }
                        }
                    }

                    if (!inShadow && (fLightProjection > 0.0f))
                    {

                        float lambert = fLightProjection;

                        fLightContrib += lambert * fLightIntensity  
                            * ((1.0f - noiseCoef) * currentMatChannel.diffuse + noiseCoef * currentMatChannel.diffuse2);

                        // Blinn 
                        // La direction de Blinn est exactement  mi chemin entre le rayon
                        // lumineux et le rayon de vue. 
                        // On calcule le vecteur de Blinn et on le rend unitaire
                        // puis on calcule le cofficient de blinn
                        // qui est la contribution spculaire de la lumire courante.

                        vecteur blinnDir = lightRay.dir - viewRay.dir;
                        {
                            float temp = blinnDir * blinnDir;
                            if (temp != 0.0f )
                            {
                                float blinn = invsqrtf(temp) * max(fLightProjection - fViewProjection , 0.0f);
                                blinn = currentMatChannel.specular * powf(blinn, currentMatChannel.power);
                                fLightContrib += blinn * fLightIntensity;
                            }
                        }
                    }
                }
            }
           
            if(fLightContrib > 0.0f)
            {
                output += fLightContrib * coef;
            }
            coef = 0.0f ;
        }
        viewRay = outRay;

        level++;
    } while ((coef > 0.0f) && (level < 10));  

    if (coef > 0.0f)
    {
        output += coef * readCubemap(myScene.cm, viewRay).getChannel(myContext.offset);
    }
    return output;
}

float AutoExposure(scene &myScene, photonmap photonMapTest[3])
{
    #define ACCUMULATION_SIZE 256
	float exposure = -1.0f;
	float accufacteur = float(max(myScene.sizex, myScene.sizey));
    float projectionDistance = 0.0f;

    if ((myScene.persp.type == perspective::conic) && myScene.persp.FOV > 0.0f && myScene.persp.FOV < 189.0f)
    {
        // On triangule avec l'angle du FOV divis par deux et 
        // la taille du viewport divis par deux
        projectionDistance = 0.5f * myScene.sizex / tanf (float(PIOVER180) * 0.5f * myScene.persp.FOV);
    }

    accufacteur = accufacteur / ACCUMULATION_SIZE;
    color mediumPoint = {0.0f, 0.0f, 0.0f};
    const float mediumPointWeight = 1.0f / (ACCUMULATION_SIZE*ACCUMULATION_SIZE);
    for (int y = 0; y < ACCUMULATION_SIZE; ++y)
    for (int x = 0 ; x < ACCUMULATION_SIZE; ++x) 
    {
        for(int offset = 0; offset < color::OFFSET_MAX; offset++ )
        {
            if (myScene.persp.type == perspective::orthogonal || projectionDistance == 0.0f)
            {
                ray viewRay = { {float(x)*accufacteur, float(y) * accufacteur, -10000.0f}, { 0.0f, 0.0f, 1.0f}};
                context myContext = context::getDefaultAir();
                myContext.offset = color::OFFSET(offset);
                float currentColor = addRay (viewRay, myScene, myContext, photonMapTest[offset]);
                mediumPoint.getChannel(color::OFFSET(offset)) += mediumPointWeight * (currentColor * currentColor);
            }
            else
            {
                vecteur dir = {(float(x)*accufacteur - 0.5f * myScene.sizex) / projectionDistance, 
                                (float(y) * accufacteur - 0.5f * myScene.sizey) / projectionDistance, 
                                1.0f}; 

                float norm = dir * dir;
                // je ne pense pas que a puisse arriver mais on n'est jamais trop prudent
                if (norm == 0.0f) 
                    break;
                dir = invsqrtf(norm) * dir;

                ray viewRay = { {0.5f * myScene.sizex,  0.5f * myScene.sizey, 0.0f}, {dir.x, dir.y, dir.z} };
                context myContext = context::getDefaultAir();
                myContext.offset = color::OFFSET(offset);
			    float currentColor = addRay (viewRay, myScene, myContext, photonMapTest[offset]);
			    mediumPoint.getChannel(color::OFFSET(offset)) += mediumPointWeight * (currentColor * currentColor);
                //mediumPoint.getChannel(color::OFFSET(offset)) = max(mediumPoint.getChannel(color::OFFSET(offset)), currentColor);
            }
        }
    }
	
    // Code d'autoexposition, on place le carr moyen  une valeur de gris intermdiaire

    float mediumLuminance = sqrtf(0.2126f * mediumPoint.red + 0.715160f * mediumPoint.green + 0.072169f * mediumPoint.blue);
//    float mediumLuminance = 0.2126f * mediumPoint.red + 0.715160f * mediumPoint.green + 0.072169f * mediumPoint.blue;
    if (mediumLuminance > 0.0f)
    {
        exposure = logf(0.6f) / mediumLuminance;
//        exposure = - 1.0f / mediumLuminance;
    }

    return exposure;
}

struct sampleDesc {float shiftx; float shifty; float sampleweight;};

const int sampleHighCount = 49; 

sampleDesc sampleHighArray[sampleHighCount] = 
{
    // poids et positions des chantillons, prcalculs. 49 chantillons par pixels, la couverture de chaque pixel
    // overlappe un peu son voisin ce qui peut rajouter du flou mais limite le risque d'aliasing 
    // (par rapport  si on choisit un box filtre exactement align sur les pixels de l'cran)
    {-0.89f,-1.11f, 0.00137f}, {-0.56f,-1.00f, 0.00412f}, {-0.22f,-0.89f, 0.00823f}, { 0.11f,-0.11f, 0.00960f}, { 0.44f,-1.00f, 0.00823f}, { 0.78f,-0.89f, 0.00412f}, { 1.11f,-1.11f, 0.00137f},
    {-1.00f,-0.78f, 0.00412f}, {-0.67f,-0.67f, 0.01235f}, {-0.33f,-0.56f, 0.02469f}, { 0.00f, 0.22f, 0.02881f}, { 0.33f,-0.67f, 0.02469f}, { 0.67f,-0.56f, 0.01235f}, { 1.00f,-0.78f, 0.00412f},
    {-1.11f,-0.44f, 0.00823f}, {-0.78f,-0.33f, 0.02469f}, {-0.44f,-0.22f, 0.04938f}, {-0.11f,-0.44f, 0.05761f}, { 0.22f,-0.33f, 0.04938f}, { 0.56f,-0.22f, 0.02469f}, { 0.89f,-0.44f, 0.00823f},
    {-0.89f,-0.11f, 0.00960f}, {-0.56f, 0.00f, 0.02881f}, {-0.22f, 0.11f, 0.05761f}, { 0.11f,-0.11f, 0.06722f}, { 0.44f, 0.00f, 0.05761f}, { 0.78f, 0.11f, 0.02881f}, { 1.11f,-0.11f, 0.00960f},
    {-1.00f, 0.22f, 0.00823f}, {-0.67f, 0.33f, 0.02469f}, {-0.33f, 0.44f, 0.04938f}, { 0.00f, 0.22f, 0.05761f}, { 0.33f, 0.33f, 0.04938f}, { 0.67f, 0.44f, 0.02469f}, { 1.00f, 0.22f, 0.00823f},
    {-1.11f, 0.56f, 0.00412f}, {-0.78f, 0.67f, 0.01235f}, {-0.44f, 0.78f, 0.02469f}, {-0.11f,-0.44f, 0.02881f}, { 0.22f, 0.67f, 0.02469f}, { 0.56f, 0.78f, 0.01235f}, { 0.89f, 0.56f, 0.00412f},
    {-0.89f, 0.89f, 0.00137f}, {-0.56f, 1.00f, 0.00412f}, {-0.22f, 1.11f, 0.00823f}, { 0.11f,-0.11f, 0.00960f}, { 0.44f, 1.00f, 0.00823f}, { 0.78f, 1.11f, 0.00412f}, { 1.11f, 0.89f, 0.00137f},
};

const int sampleMediumCount = 16; 

sampleDesc sampleMediumArray[sampleMediumCount] = 
{
    // poids et positions des chantillons, prcalculs. 16 chantillons par pixels, la couverture de chaque pixel
    // overlappe un peu son voisin ce qui peut rajouter du flou mais limite le risque d'aliasing 
    // (par rapport  si on choisit un box filtre exactement align sur les pixels de l'cran)
    {-0.675f, -0.875f, 0.015625f}, {-0.125f, -0.675f, 0.046875f}, {0.375f, -0.875f, 0.046875f}, {0.875f, -0.675f, 0.015625f}, 
    {-0.875f, -0.375f, 0.046875f}, {-0.375f, -0.125f, 0.140625f}, {0.125f, -0.375f, 0.140625f}, {0.675f, -0.125f, 0.046875f}, 
    {-0.675f,  0.125f, 0.046875f}, {-0.125f,  0.375f, 0.140625f}, {0.375f,  0.125f, 0.140625f}, {0.875f,  0.375f, 0.046875f}, 
    {-0.875f,  0.675f, 0.015625f}, {-0.375f,  0.875f, 0.046875f}, {0.125f,  0.675f, 0.046875f}, {0.675f,  0.875f, 0.015625f}, 
};

const int sampleLowCount = 4; 

sampleDesc sampleLowArray[sampleLowCount] = 
{
    // antialiasing classique  4 samples par pixels. Les chantillons ne sont pas aligns pour 
    // amliorer le lissage sur les bords.
    {-0.375f, -0.125f, 0.25f}, {0.125f, -0.375f, 0.25f}, {-0.125f,  0.375f, 0.25f}, {0.375f,  0.125f, 0.25f}, 
};

bool draw(char* outputName, scene &myScene)
{
    int x, y;
    ofstream imageFile(outputName,ios_base::binary);
    if (!imageFile)
        return false;
    // specific TGA handling code
    imageFile.put(0).put(0);
    imageFile.put(2);                  /* uncompressed RGB */

    imageFile.put(0).put(0);
    imageFile.put(0).put(0);
    imageFile.put(0);

    imageFile.put(0).put(0);           /* X origin */
    imageFile.put(0).put(0);           /* y origin */

    imageFile.put((myScene.sizex & 0x00FF)).put((myScene.sizex & 0xFF00) / 256);
    imageFile.put((myScene.sizey & 0x00FF)).put((myScene.sizey & 0xFF00) / 256);
    imageFile.put(24);                 /* 24 bit bitmap */
    imageFile.put(0);
	
    photonmap channelPhotonMap[3];

    if (myScene.lgtTab.size() == 0)
        return false;
    {
        const int maxPhotonStored = myScene.phMapProp.size;
        if (maxPhotonStored)
        {
            context myContext = context::getDefaultAir();
            channelPhotonMap[color::OFFSET_RED].reserve( maxPhotonStored );
            channelPhotonMap[color::OFFSET_GREEN].reserve( maxPhotonStored );
            channelPhotonMap[color::OFFSET_BLUE].reserve( maxPhotonStored );
            for (int offset =0 ; offset < 3 ; offset ++)
            {
                int nPhotonSent = 0;
                for (;;)
                {
                    ray viewRay;        
                    do
                    {
                        viewRay.dir.x = (2.0f / RAND_MAX) * rand() - 1.0f;
                        viewRay.dir.y = (2.0f / RAND_MAX) * rand() - 1.0f;
                        viewRay.dir.z = (2.0f / RAND_MAX) * rand() - 1.0f;
                    } 
                    while  (viewRay.dir.x *viewRay.dir.x + viewRay.dir.y * viewRay.dir.y + viewRay.dir.z * viewRay.dir.z > 1.0f );

                    float temp = viewRay.dir * viewRay.dir;
	                if (temp == 0.0f)
		                continue;
                    viewRay.dir = invsqrtf(temp) * viewRay.dir;

                    int lightNumber = rand() % int(myScene.lgtTab.size());
                    viewRay.start = myScene.lgtTab[lightNumber].pos;
                    myContext.offset = color::OFFSET(offset);
                    if (!addPhotonRay(viewRay, myScene, myContext, channelPhotonMap[offset]))
                    {
                        // on a atteint la limite du stockage de photons, on peut arreter d'en ajouter
                        break;
                    };
                    nPhotonSent++;
                }
                float fTotalLightIntensity = 0.0f;
                for (int iLight = 0; iLight < int(myScene.lgtTab.size()); ++iLight)
                {
                    fTotalLightIntensity += myScene.lgtTab[iLight].intensity.getChannel(color::OFFSET(offset));
                }
                float fWattagePerPhoton = 4 * 3.1415927f * fTotalLightIntensity / nPhotonSent;
                channelPhotonMap[offset].setWattagePerPhoton(fWattagePerPhoton);
            }
        }
    }


    float exposure = AutoExposure(myScene, channelPhotonMap);
    // on fait des maths ici parce que nos paramtres sont
    // implicites, on donne le FOV horizontal (le champ de vision)
    // en paramtre et il faut en dduire  quelle distance 
    // virtuelle se trouve notre plan de projection du point d'origine. 
    float projectionDistance = 0.0f;

    if ((myScene.persp.type == perspective::conic) && myScene.persp.FOV > 0.0f && myScene.persp.FOV < 189.0f)
    {
        // On triangule avec l'angle du FOV divis par deux et 
        // la taille du viewport divis par deux
        projectionDistance = 0.5f * myScene.sizex / tanf (float(PIOVER180) * 0.5f * myScene.persp.FOV);
    }

    for (y = myScene.sizey - 1; y >= 0 ; --y)
    {
        for (x = 0 ; x < myScene.sizex; ++x)
        {
            color output = {0.0f, 0.0f, 0.0f};
            for(int offset = 0; offset < color::OFFSET_MAX; offset++ )
            {
                float temp;
                int sampleCount;
                sampleDesc *sampleArray;

                switch(myScene.supersampleType)
                {
                case scene::high:
                    sampleCount = sampleHighCount;
                    sampleArray = sampleHighArray;
                    break;
                case scene::medium:
                    sampleCount = sampleMediumCount;
                    sampleArray = sampleMediumArray;
                    break;
                case scene::low:
                default:
                    sampleCount = sampleLowCount;
                    sampleArray = sampleLowArray;
                    break;
                }

                for (int sample = 0; sample < sampleCount; ++sample)
                {
                    const float fragmentx = float(x) + sampleArray[sample].shiftx;
                    const float fragmenty = float(y) + sampleArray[sample].shifty;
                    const float sampleRatio = sampleArray[sample].sampleweight;
                    temp = 0.0f;
                    float fTotalWeight = 0.0f;

                    if (myScene.persp.type == perspective::orthogonal || projectionDistance == 0.0f)
                    {
                        ray viewRay = { {fragmentx, fragmenty, -10000.0f}, { 0.0f, 0.0f, 1.0f}};
                        for (int i = 0; i < myScene.complexity; ++i)
                        {                  
                            context myContext = context::getDefaultAir();
                            myContext.offset = color::OFFSET(offset);
                            float rayResult = addRay (viewRay, myScene, myContext, channelPhotonMap[offset]);
                            fTotalWeight += 1.0f; 
                            temp += rayResult;
                        }
                    }
                    else
                    {
                        vecteur dir = {(fragmentx - 0.5f * myScene.sizex) / projectionDistance, 
                                    (fragmenty - 0.5f * myScene.sizey) / projectionDistance, 
                                    1.0f}; 

                        float norm = dir * dir;
                        // je ne pense pas que a puisse arriver mais on n'est jamais trop prudent
                        if (norm == 0.0f) 
	                        break;
                        dir = invsqrtf(norm) * dir;
                        // le point de dpart est toujours le centre optique de la camra
                        // il peut tre perturb plus tard pour un effet de profondeur de champ
                        point start = {0.5f * myScene.sizex,  0.5f * myScene.sizey, 0.0f};
                        // le point vis est le point invariant du pixel courant,
                        // c'est  dire que par construction tous les rayons qui participent
                        // au pixel courant doivent passer par ce point
                        point ptVise = start + myScene.persp.clearPoint * dir;

                        for (int i = 0; i < myScene.complexity; ++i)
                        {                  
                            ray viewRay = { {start.x, start.y, start.z}, {dir.x, dir.y, dir.z} };

                            if (myScene.persp.dispersion != 0.0f)
                            {
                                vecteur vDisturbance;                        
                                vDisturbance.x = (myScene.persp.dispersion / RAND_MAX) * (1.0f * rand());
                                vDisturbance.y = (myScene.persp.dispersion / RAND_MAX) * (1.0f * rand());
                                vDisturbance.z = 0.0f;

                                viewRay.start = viewRay.start + vDisturbance;
                                viewRay.dir = ptVise - viewRay.start;
                                
                                norm = viewRay.dir * viewRay.dir;
                                if (norm == 0.0f)
                                    break;
                                viewRay.dir = invsqrtf(norm) * viewRay.dir;
                            }
                            context myContext = context::getDefaultAir();
                            myContext.offset = color::OFFSET(offset);
                            float rayResult = addRay (viewRay, myScene, myContext, channelPhotonMap[offset]);
                            fTotalWeight += 1.0f;
                            temp += rayResult;
                        }
                    }

                    temp /= fTotalWeight;
                    // exposition photo
                    temp = (1.0f - expf(temp * exposure));
                    output.getChannel(color::OFFSET(offset) ) += sampleRatio * temp; 
                }
            }
						
            // gamma correction
            
            float invgamma = 0.45f; // c'est la valeur fixe par le standard sRGB = 1 / 2.2
            output.blue  = powf(output.blue, invgamma);
            output.red   = powf(output.red, invgamma);
            output.green = powf(output.green, invgamma);
            
            if (y < 10)
            {
                if ((x / 10) & 1)
                {
                    imageFile.put((unsigned char)186).put((unsigned char)186).put((unsigned char)186);
                }
                else if ( y & 1)
                {
                    imageFile.put((unsigned char)255).put((unsigned char)255).put((unsigned char)255);
                }
                else
                {
                    imageFile.put((unsigned char)0).put((unsigned char)0).put((unsigned char)0);
                }
            }
            else
            {
                imageFile.put((unsigned char)min(output.blue*255.0f,255.0f)).put((unsigned char)min(output.green*255.0f, 255.0f)).put((unsigned char)min(output.red*255.0f, 255.0f));
            }
        }
	}
	return true;
}

int __cdecl main(int argc, char* argv[])
{
    if (argc < 3)
    {
        cout << "Usage : Raytrace.exe Scene.txt Output.tga" << endl;
        return -1;
    }
    scene myScene;
    if (!init(argv[1], myScene))
    {
        cout << "Echec  l'ouverture du fichier scene." << endl;
        return -1;
    }
    if (!draw(argv[2], myScene))
    {
        cout << "Echec  la cration du fichier image." << endl;
        return -1;
    }
    return 0;
}
