#include "photonmap.h"
#include "photontrace.h"
#include <vector>
#include <limits>
#include <algorithm>
#include <cmath>
#include <iostream>
using namespace std;

namespace
{
    struct kdnode{
        enum {
            dimensionX = 0,
            dimensionY = 1,
            dimensionZ = 2
        } dimensionPlane;
        // float fCoordinatePlane;
        int nPhotonMedian;
    };
}

class photoncache 
{
public:
    vector<photon> photonTab;
    vector<vecteur> normalTab;
    vector<kdnode> nodeTab;
    void balance(vector<int>& tab, int p);
    bool bBalanced;
};

template<int n> struct sortingCriterion
{
    const photon *photonTab; 
    bool operator() (int p1, int p2) 
    {
        if (n == kdnode::dimensionX)
        {
            return photonTab[p1].pos.x < photonTab[p2].pos.x;
        }
        
        if (n == kdnode::dimensionY)
        {
            return photonTab[p1].pos.y < photonTab[p2].pos.y;
        }
        
        if (n == kdnode::dimensionZ)
        {
            return photonTab[p1].pos.z < photonTab[p2].pos.z;
        }
    }
};

void photoncache::balance(vector<int>& tab, int p)
{
    if (tab.size() == 0)
        return;
    vector<int> tabLess;
    vector<int> tabMore;
    if (2 * p + 1 < 100 * int(photonTab.size()))
    {
        kdnode node;
        node.nPhotonMedian = -1;
        if (tab.size() == 1)
        {
            node.nPhotonMedian = tab[0];
        }
        else
        {
            float minX, minY, minZ;
            float maxX, maxY, maxZ;

            minX = minY = minZ = numeric_limits<float>::max();
            maxX = maxY = maxZ = - numeric_limits<float>::max();
            tabLess.reserve(tab.size());
            tabMore.reserve(tab.size());
            for (int i = 0; i < int(tab.size()); ++i)
            {
                minX = min(minX, photonTab[tab[i]].pos.x);
                minY = min(minY, photonTab[tab[i]].pos.y);
                minZ = min(minZ, photonTab[tab[i]].pos.z);
                maxX = max(maxX, photonTab[tab[i]].pos.x);
                maxY = max(maxY, photonTab[tab[i]].pos.y);
                maxZ = max(maxZ, photonTab[tab[i]].pos.z);
            }

            if ( maxX - minX >= maxY - minY )
            {
                if (maxX - minX >= maxZ - minZ)
                {
                    node.dimensionPlane = kdnode::dimensionX;
                    sortingCriterion<kdnode::dimensionX> pred = {&photonTab[0]};
                    sort(tab.begin(), tab.end(), pred);
                    node.nPhotonMedian = tab[int(tab.size()) / 2]; 
                    for (int i = 0; i< int(tab.size()) / 2; ++i)
                    {
                        tabLess.push_back(tab[i]);
                    } 
                    for (int i = int(tab.size()) / 2 + 1; i< int(tab.size()); ++i)
                    {
                        tabMore.push_back(tab[i]);
                    }
                }
                else 
                {
                    node.dimensionPlane = kdnode::dimensionZ;
                    sortingCriterion<kdnode::dimensionZ> pred = {&photonTab[0]};
                    sort(tab.begin(), tab.end(), pred);
                    node.nPhotonMedian = tab[int(tab.size()) / 2]; 
                    for (int i = 0; i< int(tab.size()) / 2; ++i)
                    {
                        tabLess.push_back(tab[i]);
                    } 
                    for (int i = int(tab.size()) / 2 + 1; i< int(tab.size()); ++i)
                    {
                        tabMore.push_back(tab[i]);
                    }
                }
            }
            else 
            {
                if (maxY - minY >= maxZ - minZ)
                {
                    node.dimensionPlane = kdnode::dimensionY;
                    sortingCriterion<kdnode::dimensionY> pred = {&photonTab[0]};
                    sort(tab.begin(), tab.end(), pred);
                    node.nPhotonMedian = tab[int(tab.size()) / 2]; 
                    for (int i = 0; i< int(tab.size()) / 2; ++i)
                    {
                        tabLess.push_back(tab[i]);
                    } 
                    for (int i = int(tab.size()) / 2 + 1; i< int(tab.size()); ++i)
                    {
                        tabMore.push_back(tab[i]);
                    }
                }
                else 
                {
                    node.dimensionPlane = kdnode::dimensionZ;
                    sortingCriterion<kdnode::dimensionZ> pred = {&photonTab[0]};
                    sort(tab.begin(), tab.end(), pred);
                    node.nPhotonMedian = tab[int(tab.size()) / 2]; 
                    for (int i = 0; i< int(tab.size()) / 2; ++i)
                    {
                        tabLess.push_back(tab[i]);
                    } 
                    for (int i = int(tab.size()) / 2 + 1; i< int(tab.size()); ++i)
                    {
                        tabMore.push_back(tab[i]);
                    }
                }
            }
        }
        if (p >= int(nodeTab.size()) )
            nodeTab.resize(p + 1); 
        nodeTab[p] = node;
    }
    balance(tabLess, 2 * p);
    balance(tabMore, 2 * p + 1);
}

bool photonmap::reserve(int myNumber)
{
    if (!container)
    {
        container = new photoncache;
    }
    if (!container)
        return false;
    container->photonTab.reserve(myNumber);
    container->normalTab.reserve(myNumber);
    number = myNumber;
    container->bBalanced = false;
    return true;
}

void photonmap::locatePhotons (const point& pos, const vecteur &normal, float fMinCos, int nPhotons, float & maxDistSquare, vector<pair<float, int> > & result, int firstNode)
{
    const vecteur *normalTab = &container->normalTab[0];
    const photon *photonTab = &container->photonTab[0];

    if (2 * firstNode + 1 < int(container->nodeTab.size()))  
    {
        int nPhotonMedian = container->nodeTab[firstNode].nPhotonMedian ;
        if (nPhotonMedian >= 0)
        {
            // Il y a un photon  ce noeud, evalue la distance au point d'intersection
            // et stocke le s'il y a lieu
            const photon * currentPhoton = &container->photonTab[nPhotonMedian];

            {
                vecteur vDistToPhoton = pos - currentPhoton->pos;
                float fDistSquare = vDistToPhoton * vDistToPhoton;

                // On ne rajoute pas un photon qui se retrouve dj plus loin que la distance max
                if ( maxDistSquare > fDistSquare)
                {
                    // On limine les photons qui ont t stocks sur une surface qui a peu de probabilit
                    // d'tre la mme que celle du point d'intersection. Sinon il y a risque de pollution entre
                    // deux surfaces proches (angle de deux murs par exemple)
                    if ( (normalTab[nPhotonMedian] * normal > fMinCos) && (photonTab[nPhotonMedian].dir * normal < 0.0f) )
                    {
                        if (int(result.size()) == nPhotons ) 
                        {
                            // le conteneur est dj plein. Mais a ne veut pas dire que
                            // le nouveau photon n'est pas plus proche que les n autres
                            // on compare les distance et on stocke le cas chant 
                            // en dcalant les autres.
                        
                            maxDistSquare = 0;
                            for (int i = 0; i < nPhotons; ++i)
                            {
                                if(result[i].first > fDistSquare)
                                {
                                    float fTemp = fDistSquare;
                                    fDistSquare = result[i].first;
                                    result[i].first = fTemp;
                                    int temp = nPhotonMedian;
                                    nPhotonMedian = result[i].second;
                                    result[i].second = temp;
                                }

                                if (result[i].first > maxDistSquare)
                                {
                                    maxDistSquare = result[i].first;
                                }
                            }
                        }
                        else 
                        {
                            // On ajoute  la liste qui n'est pas encore pleine
                            result.push_back(pair<float, int>(fDistSquare,nPhotonMedian));
                        }
                    }
                }
            }

            float fDist;
            switch (container->nodeTab[firstNode].dimensionPlane)
            {
            case kdnode::dimensionX:
                fDist = pos.x - currentPhoton->pos.x;
                break;
            case kdnode::dimensionY:
                fDist = pos.y - currentPhoton->pos.y;
                break;
            case kdnode::dimensionZ:
            default:
                fDist = pos.z - currentPhoton->pos.z;
                break;
            }
            if (fDist <= 0.0f)
            {
                float fDistSquare = fDist * fDist;
                locatePhotons(pos, normal, fMinCos, nPhotons, maxDistSquare, result, 2 * firstNode);
                if ( maxDistSquare >= fDistSquare )
                {
                    // Tout l'intrt du KD-Tree rside dans le pruning
                    // si l'on a dj trouv n photons, pas la peine de chercher dans la liste
                    // des photons qui se trouvent plus loin que la distance max.
                    locatePhotons(pos, normal, fMinCos, nPhotons, maxDistSquare, result, 2 * firstNode + 1);
                }
            }
            else
            {
                float fDistSquare = fDist * fDist;
                locatePhotons(pos, normal, fMinCos, nPhotons, maxDistSquare, result, 2 * firstNode + 1);
                if ( maxDistSquare >= fDistSquare )
                {
                    // Tout l'intrt du KD-Tree rside dans le pruning
                    // si l'on a dj trouv n photons, pas la peine de chercher dans la liste
                    // des photons qui se trouvent plus loin que la distance max.
                    locatePhotons(pos, normal, fMinCos, nPhotons, maxDistSquare, result, 2 * firstNode);
                }
            }
        }
    }
}

int photonmap::findNearestPhotons (const point& pos, const vecteur & normal, float maxAngle, int nPhotons, priorityHeap& result )
{
    if (!container)
        return 0;

    int limit = int(container->photonTab.size());

    if (!container->bBalanced)
    {
        vector<int> tab;
        tab.reserve(limit);
        for (int i = 0; i < limit; ++i )
        {
            tab.push_back(i);
        }
        container->balance(tab, 1);
        container->bBalanced = true;
    }
    float origMaxDistSquare = result.maxDistSquare;
    result.maxDistSquare = min( 2 * max(fLastMaxDistSquare, fSecondLastMaxDistance), origMaxDistSquare);
    static vector<pair<float, int> > resultTab;
    resultTab.clear();

    float fMinCos = fabsf(cosf(maxAngle));

    locatePhotons(pos, normal, fMinCos, nPhotons, result.maxDistSquare, resultTab, 1);

    fSecondLastMaxDistance = (fSecondLastMaxDistance + fLastMaxDistSquare) * 0.5f;
    fLastMaxDistSquare = result.maxDistSquare;

    // si on n'a pas trouv assez de photons, on ressaie avec une distance plus grande
    if (int(resultTab.size()) < nPhotons)
    {
        fLastMaxDistSquare = origMaxDistSquare;
    }

    for (vector<pair<float, int> >::const_iterator it = resultTab.begin(); it != resultTab.end(); ++it)
    {
        result.resultTab.push_back(pair<float, const photon*>((*it).first, &container->photonTab[(*it).second]));
    }

    return int(result.resultTab.size());
}

bool photonmap::addPhoton (const photon & newPhoton, const vecteur & newNormal)
{
    if (!container)
        return false;
    if (int(container->photonTab.size()) >= number )
        return false;
    container->photonTab.push_back(newPhoton);
    container->normalTab.push_back(newNormal);
    return true;
}

photonmap::~photonmap()
{
    delete container;
}
