// include et dfinitions diverses
#include <cmath>
#include "def.h"
#include "cubemap.h"
#include "perlin.h"
#include "photontrace.h"
#include "photonmap.h"
#include "Sphere.h"
#include "Ray.h"
#include "Scene.h"
using namespace std;

/*
 * La fonction de lancer des photons est identique ou presque
 *  celle de lancer de rayon. La diffrence est le traitement
 *  effectuer en cas d'intersection avec un matriel diffus
 * on ne calcule pas l'clairage immdiatement
 * mais on stocke le photon dans une structure hierarchis
 * pour les retrouver plus facilement dans la phase d'envoi des rayons
 * 
 */

bool addPhotonRay(ray viewRay, scene &myScene, context myContext, photonmap& myPhotonMap)
{
    int level = 0;

    do {
        point ptHitPoint;
        vecteur vNormal;
        material currentMat;
        {
            float distance;

            if (!findIntersection(/*IN*/viewRay, /*IN*/myScene, 
                                /*OUT*/ptHitPoint, /*OUT*/vNormal, /*OUT*/currentMat, /*OUT*/distance))
            {
                break;
            }

            if (myContext.absorption != 1.0f) 
            {   
                // la lumire transmise dans le mdium transparent
                // est affecte par la diffusion due aux impurets
                // la beer's law s'applique pour les faibles concentrations
                float coef = powf(myContext.absorption, distance);
                float fRoulette = (1.0f / RAND_MAX) * rand();
                if (fRoulette > coef )
                {
                    // Le photon s'est retrouv absorb avant d'arriver au point d'intersection.
                    // On peut juste s'arreter de stocker des photons
                    break; 
                }
            }
        }

        bool bInside;
        if (vNormal * viewRay.dir > 0.0f)
        {
            vNormal = -1.0f * vNormal;
            bInside = true;
        }
        else
        {
            bInside = false;
        }

        // perturbation du vecteur normal par un bruit alatoire 
        // on le renormalise derrire pour le conserver unitaire
        // pour les calculs d'clairage
        if (currentMat.bump != 0.0f)
        {
            float noiseCoefx = noisef(0.1f * ptHitPoint.x, 0.1f * ptHitPoint.y,0.1f * ptHitPoint.z);
            float noiseCoefy = noisef(0.1f * ptHitPoint.y, 0.1f * ptHitPoint.z,0.1f * ptHitPoint.x);
            float noiseCoefz = noisef(0.1f * ptHitPoint.z, 0.1f * ptHitPoint.x,0.1f * ptHitPoint.y);

            vNormal.x = (1.0f - currentMat.bump ) * vNormal.x + currentMat.bump * noiseCoefx; 
            vNormal.y = (1.0f - currentMat.bump ) * vNormal.y + currentMat.bump * noiseCoefy; 
            vNormal.z = (1.0f - currentMat.bump ) * vNormal.z + currentMat.bump * noiseCoefz; 
            float temp = vNormal * vNormal;
            if (temp == 0.0f)
                break;
            vNormal = invsqrtf(temp)  * vNormal;
        }

        const materialChannel & currentMatChannel = currentMat.getChannel(myContext.offset);

        ray outRay = viewRay;

        bool bDiffuse = LookForDiffuse(viewRay, ptHitPoint, vNormal, currentMatChannel, bInside, myContext, outRay);

        if (bDiffuse)
        {
            // notre BRDF nous dit que le photon est soit rflechi soit rfract
            // soit absorb par la surface diffuse.. 
            // Pas de color bleeding d  la diffusion.. pour l'instant

            if (myScene.phMapProp.directLighting || (level > 0) )
            {
                // La surface n'est pas totalement transitive, on peut stocker le photon incident
                // on ne le fait que pour les niveaux suprieurs  zero si l'on dcide
                // d'utiliser les formules classiques pour l'clairage direct
                photon currentPhoton; // {ptHitPoint, viewRay.dir, coef}; 
                currentPhoton.pos = ptHitPoint;
                currentPhoton.dir = viewRay.dir;

                if (!myPhotonMap.addPhoton(currentPhoton, vNormal)) 
                {
                    // Le container a photon a atteint ses limites.
                    // On peut juste s'arreter de stocker des photons
                    return false;   
                };
            }
            return true;
        }

        viewRay = outRay;

        level++;
    } while (level < 10);  
    return true;
}

